from ipdb import set_trace as idebug

import PyQt5.QtWidgets as QtWidget
import PyQt5.QtCore as QtCore
import PyQt5.QtGui as QtGui
import PyQt5.QtWidgets

"""
A quick widget that can be called from ipython to display a dataframe
"""

class QTable(PyQt5.QtWidgets.QDialog):
    def __init__(self, df, num=1000, title="Dataframe", parent=None):
        super().__init__(parent)
        app = PyQt5.QtWidgets.QApplication.instance()
        if app is None:
            app = PyQt5.QtWidgets.QApplication([])

        self.create_layout(df, num, title)
        self.selector = ColumnSelector(self.table)
        self.selector.hide()

        #I think I have to subclass QTableWidget and override keyReleaseEvent
        self.keyReleaseEvent = self.process_key_press

    def create_layout(self, df, num, title):
        self.button = QtWidget.QPushButton("Show/hide Columns")
        self.button.clicked.connect(self.toggle_button)
        self.table = TableWidget(df, num=num)


        layout = QtWidget.QVBoxLayout()
        layout.addWidget(self.button)
        layout.addWidget(self.table)
        self.setLayout(layout)

        self.resize(self.table.width(), self.table.height())
        self.setMaximumSize(self.table.getMaxWidth(), 10000)

        self.setWindowTitle(title)

    def process_key_press(self, eventQKeyEvent):
        key = eventQKeyEvent.key()
        if key == 81:  #The letter [q]
            self.selector.hide()
            self.hide()
            self.close()

    def toggle_button(self):
        if self.selector.isVisible():
            self.selector.hide()
        else:
            self.selector.show()

    def toggleColumn(self, sender_label, state):
        cols = self.table.df.columns
        for i in range(self.ncol):
            if cols[i] == sender_label:
                if state:
                    self.table.showColumn(i)
                else:
                    self.table.hideColumn(i)

    def showAll(self):
        for i in range(self.ncol):
            self.table.showColumn(i)

    def hideAll(self):
        for i in range(self.ncol):
            self.table.hideColumn(i)


class TableWidget(PyQt5.QtWidgets.QTableWidget):
    def __init__(self, df, num=1000, parent=None):
        self.df = df[:num]
        self.nrow = len(self.df)
        self.ncol = len(self.df.columns)
        super().__init__(self.nrow, self.ncol, parent=parent)
        self.max_width = 1000

        # PyQt5.QtWidgets.QTableWidget(self.nrow, self.ncol)
        self.set_table_elements(self.df)
        self.set_size_policy()
        self.draw_row_guides()
        self.setSortingEnabled(True)

        header = self.horizontalHeader()
        header.sectionClicked.connect(self.draw_row_guides)
        self.resizeRowsToContents()

    def set_table_elements(self, df):
        self.setHorizontalHeaderLabels(df.columns)

        QItem = PyQt5.QtWidgets.QTableWidgetItem  #Mnumonic
        for i, key in enumerate(df.columns):
            col = df[key]
            for j, elt in enumerate(col):
                # print(i, j, elt)
                if elt is None:
                    item = QItem(" ")
                    continue
                try:
                    item = QItem("%g"% (elt))
                except TypeError:
                    item = QItem(str(elt))
                # Mark item as readonly
                item.setFlags( item.flags() & ~QtCore.Qt.EditRole)
                self.setItem(j, i, item)

        self.resizeColumnsToContents()
        self.resizeRowsToContents()

    def set_size_policy(self):
        width_pix = self.horizontalHeader().length() + self.verticalHeader().width() + 20
        height_pix = self.verticalHeader().length() + self.horizontalHeader().width()
        self.setMaximumSize(width_pix, height_pix)
        self.max_width = width_pix

        width_pix = min(width_pix, 1000)
        height_pix = min(height_pix, 1000)
        self.resize(width_pix, height_pix)

    def getMaxWidth(self):
        return self.max_width

    def draw_row_guides(self):
        """This is called by the HeaderView.isClicked signal"""
        guideStep = 5

        for i in range(self.nrow):
            clr = QtGui.QColor('#FFFFFF')
            if i % guideStep == guideStep -1:
                clr = QtGui.QColor('#DDDDFF')

            for j in range(self.ncol):
                item = self.item(i, j)
                if item is not None:
                    item.setBackground(clr)

    def toggleColumn(self, sender_label, state):
        cols = self.df.columns
        for i in range(self.ncol):
            if cols[i] == sender_label:
                if state:
                    self.showColumn(i)
                else:
                    self.hideColumn(i)

    def showAll(self):
        for i in range(self.ncol):
            self.showColumn(i)

    def hideAll(self):
        for i in range(self.ncol):
            self.hideColumn(i)


class ColumnSelector(PyQt5.QtWidgets.QDialog):
    def __init__(self, table, parent=None):
        self.table = table
        df = table.df
        PyQt5.QtWidgets.QDialog.__init__(self, parent)
        self.setMinimumWidth(150)

        layout = PyQt5.QtWidgets.QVBoxLayout(self)
        button1 = QtWidget.QPushButton("Select All")
        button1.clicked.connect(self.showAll)
        button2 = QtWidget.QPushButton("Hide All")
        button2.clicked.connect(self.hideAll)
        layout.addWidget(button1)
        layout.addWidget(button2)
        self.setWindowTitle("Select Columns")

        self.boxes = []
        for i, col in enumerate(df.columns):
            print(i, col)
            checkbox = QtWidget.QCheckBox(col)
            checkbox.setChecked(True)
            checkbox.stateChanged.connect(self.onToggle)
            layout.addWidget(checkbox)
            self.boxes.append(checkbox)
        self.layout = layout
        self.show()

        self.keyReleaseEvent = self.quit

    def quit(self, eventQKeyEvent):
        key = eventQKeyEvent.key()
        if key == 81:  #The letter [q]
            self.hide()

    def showAll(self):
        self.table.showAll()  # Maybe unecessary?

        # Update all the checkboxes
        for box in self.boxes:
            box.setChecked(True)

    def hideAll(self):
        self.table.hideAll()  # Maybe unecessary?

        # Update all the checkboxes
        for box in self.boxes:
            box.setChecked(False)

    def onToggle(self, state):
        sender = self.sender()
        self.table.toggleColumn(sender.text(), state > 0)


if __name__ == "__main__":
    import pandas as pd
    import sys

    if len(sys.argv) != 2:
        print("Usage: qtable.py file.csv")
        sys.exit(1)

    app = PyQt5.QtWidgets.QApplication([])
    df = pd.read_csv(sys.argv[1])
    print("loading data")
    tab = QTable(df)
    tab.show()
    sys.exit(app.exec())




from  ipdb import set_trace as idebug 
import matplotlib.pyplot as plt
import numpy as np
import re



class PlotWheel(object):
    """An interactive plot
    Clicking in the plot activates a requested function which can
    interact with the data.

    This is a base class extend it with a class that implements a 
    plotting function
    """
    def __init__(self, index_max, i0=0):
        #Set input arguments as class variables.
        self.index = i0
        self.index_max = index_max
        
        f = plt.gcf()
        self.draw_plot(self.index)

        self.updateEvent = 'key_press_event'
        f.canvas.mpl_disconnect(self.updateEvent)
        f.canvas.mpl_connect(self.updateEvent, self)


    def __del__(self):
        self.disconnect()

    def __call__(self, event):
        request = event.key

        if request == 'n':
            if self.index == self.index_max - 1:
                print("Already at end")
                return 
            self.index += 1
        elif request == 'b':
            if self.index == 0:
                print("Already at start")
                return 
            self.index -= 1
        elif request == 'N':
            if self.index == self.index_max - 1:
                print("Already at end")
                return
            step = min(10, self.index_max - self.index - 1)
            self.index += step
        elif request == 'B':
            if self.index == 0:
                print("Already at start")
                return
            step = min(10, self.index)
            self.index -= step
        elif request == 'r':
            self.index = int(self.index_max * np.random.rand())
        elif request == 'S':
            fn = self.get_savefile_name(self.index)
            plt.savefig(fn)
            print("Saved to %s" %(fn))
        elif request == 'q':
            return

        self.draw_plot(self.index) 

    def get_savefile_name(self, i):
        return "frame%04i.png" %(i)

    def disconnect(self):
        """Disconnect the figure from the interactive object

        This method is called by the destructor, but Python
        doesn't always call the destructor, so you can do so
        explicitly.         And probl should
        """
        print ("Disconnecting...")
        plt.gcf().canvas.mpl_disconnect(self.updateEvent)

    def draw_plot(self, i):
        """Over ride this method to do setup/tear down for each plot. 

        See GroupPlotWheel for an example
        """
        return self.plot(i)

    def plot(self, i):
        raise NotImplementedError


class GroupPlotWheel(PlotWheel):
    """Base class for plotting the results of a groupby"""
    def __init__(self, df, col, i0=0):
        self.gr = df.groupby(col)
        self.keys = list(self.gr.groups.keys())

        PlotWheel.__init__(self, len(self.keys), i0=i0)

    def draw_plot(self, i):
        name = self.keys[i]
        df = self.gr.get_group(name) 
        
        plt.clf()
        self.plot(df, name, i)
        plt.pause(.01)

    def get_savefile_name(self, i):
        name = self.keys[i]
        name = re.sub(" ", "_", name)
        return "%s.png" %(name)

    def plot(self, df, name, i):
        """

        Implement this method in the child class to create the particular plot you want.

        If you need information for data outside of a single group to create your plot,
        reimpement draw_plot() instead.

        Inputs
        ---------        
        df
            (Pandas Dataframe) Data from this group
        name
            (string?) Unique identifier, or key, for this group
        i
            (int) A number between 0 and the number of groups that can be plotted 
        """
        raise NotImplementedError



from ipdb import set_trace as idebug
import matplotlib.pyplot as plt
import matplotlib as mpl
import pandas as pd
import numpy as np

import  frmplots.norm as fnorm 

"""
A cartogram is a map where the drawn sizes of geographic regions are 
based on some property of the data other than their area in square kilometres.

See https://en.wikipedia.org/wiki/Cartogram

This implementation creates maps where every geographic element
is drawn as a square of equal size. These squares are coloured by
the property you wish to plot.

Based on plotKeplerFocalPlane() from the Kepler days, but adapted
to be more general.

"""

class AbstractCartogram():
    """Extend this class to create new cartograms

    Extending the class if often as simple as setting the paramets
    of the `SHAPES` variable.

    SHAPES is a dictionary matching region names to 
    """
    SHAPES  = dict()

    def __init__(self):
        pass 

    def plot(self, names, values, **kwargs):
        cmap = kwargs.pop('cmap', plt.cm.viridis)
        vmin = kwargs.pop('vmin', np.min(values))
        vmax = kwargs.pop('vmax', np.max(values))
        norm = kwargs.pop('norm', fnorm.DiscreteNorm(7, vmin, vmax))
        fmt = kwargs.pop('fmt', "%g")

        ax = plt.gca()
        assert len(names) == len(values)
        for i in range(len(names)):
            n = names[i]
            v = values[i]

            #TODO do I need an artist collection?
            clr = cmap(norm(v))
            c, r = self.shape_lookup(n)
            patch = plt.Rectangle((c,r), .9, .8, color=clr)
            ax.add_patch(patch)

            label_square(c, r, v, fmt, clr)  #Add the value in the square
            plt.text(c+.5, r+.82, n, ha="center")  #Add placename on top

        #Remove the axis ticks
        plt.axis([0, 9, 0, 6])
        ax.xaxis.set_ticks([])
        ax.yaxis.set_ticks([])

        cb = make_colorbar(ax, cmap, norm)
        return cb

    def shape_lookup(self, name):
        shapes = self.SHAPES
        if isinstance(name, str):
            return shapes[name]
        elif isinstance(name, int):
            return list(shapes.values())[name]
        else:
            raise KeyError("Name must be int or string")


class KeplerFocalPlane(AbstractCartogram):
    pass 


class MarylandCounties(AbstractCartogram):
    SHAPES = {
        "Allegany": (1,5),
        "Anne Arundel" : (5,3),

        "Baltimore City": (5,4),
        "Baltimore County": (5,5),

        "Calvert": (5,2),
        "Caroline": (8,2),
        "Carroll": (4,5),
        "Cecil": (7,5),
        "Charles": (4,2),

        "Dorchester": (7,1),

        "Frederick": (3,5),
        "Garett": (0, 5),
        "Harford": (6,5),
        "Howard": (4,4),
        "Kent": (7,4),
        "Montgomery": (3,4),
        "Prince George's": (4,3),
        "Queen Anne's": (7,3),
        "Somerset": (7, 0),
        "St. Mary's": (5,1),
        "Talbot": (7, 2),

        "Washington": (2,5),
        "Wimcomico": (8, 1),
        "Worcester": (8,0),
    }

    def plot(self, names, values, **kwargs):
        AbstractCartogram.plot(self, names, values, **kwargs)
        plt.plot([3.5], [3.5], '*', ms=28)


class BaltimoreCountyCensusTracts(AbstractCartogram):
    pass 


class BaltimoreCountyCensusPlaces(AbstractCartogram):
    pass 


def label_square(col, row, val, fmt, clr):
    textcolor='w'
    if np.prod(clr) > .1:
        textcolor='k'

    text = fmt %(val)
    plt.text(col+.45, row+.4, text, \
        va="center", 
        ha="center", 
        bbox=None, 
        color=textcolor,
        fontsize=26,
    )


def make_colorbar(ax, cmap, norm):
    #Create a colorbar. Because we don't call scatter or imshow
    #this is more involved than usual
    cax, kw = mpl.colorbar.make_axes(ax)
    cb = mpl.colorbar.ColorbarBase(cax, cmap=cmap, norm=norm)

    #Set the main panel as the default axis again (otherwise)
    #future plotting will happen on the colourbar itself
    plt.sca(ax)
    return cb


def test_maryland():
    md = MarylandCounties()

    names = list(md.SHAPES.keys())
    values = np.arange(24)

    plt.clf()
    md.plot(names, values)


"""
Created on Mon Nov 28 14:15:53 2016
Some common plots I like to make
@author: fergal

Some tips I haven't written to functions yet
---------------------------------------------

The better way to do subplots

::
    grid = plt.gcf().subplot_mosaic("abc\def", sharex=True)
    plt.sca(grid['a'])

How to have dates on the colour bar axis

::
    from matplotlib.dates import DateFormatter,date2num
    dates = pd.date_range('2020-01-01', '2020-12-01', freq='D')
    date_num = list(map(date2num, dates))
    plt.scatter(x, y, c=date_num)
    plt.colorbar(format=DateFormatter("%Y-%m-%d"))
    
    
    
How to change the date format in the toolbar of an interactive plot

::
    import matplotlib.dates as mdate
    def foo(x):
        return mdate.num2date(x).isoformat()[:19]


    def main():
        ...
        ax = plt.gca()
        ax.fmt_xdata = foo
        
There is another function pointer called fmt_ydata for the y data format
An exercise is to figure out 

* How to check if the x axis is plotting dates or not 
* How to set this up in the rcParams

"""
import gzip

import matplotlib.collections as mcollect
from matplotlib.gridspec import GridSpec
import matplotlib.patheffects as meffect
import matplotlib.transforms as mtrans
import matplotlib.colors as mcolors
import matplotlib.ticker as mticker
import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import matplotlib as mpl
import pandas as pd
import numpy as np
import pickle
import copy

from . import norm as fnorm
from . galaxyplot import galaxyPlot
from . import plotstyle

DEFAULT_CMAP = plt.cm.viridis
add_watermark = plotstyle.add_watermark
mark_as_draft = plotstyle.mark_as_draft

def add_figure_labels(xlabel, ylabel, **kwargs):
    """Add axis labels to a figure composed of subplots.

    Inputs
    ----------
    xlabel
        (str) Label for x axis
    ylabel
        (str) Label for y axis
    
    All other options are passed to plt.xlabel and plt.ylabel.
    Use plt.suptitle to put a title over the subplots
    """

    fig = plt.gcf()
    ax = fig.gca()
    fig.add_subplot(111, frameon=False)
    plt.tick_params(labelcolor='none', top=False, bottom=False, left=False, right=False, which='both')
    plt.xlabel(xlabel, **kwargs)
    plt.ylabel(ylabel, **kwargs)  
    plt.sca(ax)


def annotate_histogram(vals, bins, **kwargs):
    """Prints the number of elements in a histogram on top of each bar.

    Slightly experimental. Only works for vertical histograms. Behaviour
    for multiple bars of the same height not well tested. 

    Example
    ------------
    ::

        bins, vals, _ = plt.hist(data)
        annotate_histogram(bins, vals, color='red', fontsize=14)

    Inputs
    ---------
    vals, bins
        The two arrays returned by `np.histogram`, or the first two values
        returned by matplotlib's `hist` command
    
    Optional Arguments
    ---------------------
    fmt
        (string) Format string for the numbers. Default is '%g'
    ha
        (string) Horizontal alignment of text. Defaults to center.
    above=True
        (bool) Whether annotations should be above or below the point 
    offset
        (float) How far above (or below) top of line to place the label.
        A value of 0 means no offsetting. A value of 1 offsets by the 
        difference between the heights of the max and in bars.
    jitter
        (float) How much to offset adjacent labels in the y direction.
        Prevents labels from over writing each other. Units the same
        as `offset`

    All other arguments are passed to `plt.text`

    """
    #Define some default values
    fmt = kwargs.pop('fmt', '%g')
    ha = kwargs.pop('ha', 'center')
    offset = kwargs.pop('offset', .02)
    offset_sign = 2* kwargs.pop('above', True) - 1
    jitter = kwargs.pop('jitter', .04)

    va = kwargs.pop('va', None)
    if va is None:
        if offset_sign > 0:
            va = 'bottom'
        else:
            va = 'top'

    if len(bins) == len(vals):
        locs = bins
    elif len(bins) == len(vals) + 1:
        locs = bins[:-1] + .5* np.diff(bins)

    scale = np.max(vals) - np.min(vals)
    offset = (offset * scale * offset_sign) + jitter * scale
    sign = -1
    old = 0
    for xpos, val in zip(locs, vals):
        if val <= 0:
            continue 

        ypos = val + offset
        if np.fabs(ypos - old) < offset:
            ypos += sign * offset 
            sign *= -1
        
        old = ypos 
        plt.text(xpos, ypos + offset, fmt%(val), ha=ha, va=va, **kwargs)


def barcode(x, clr='C0', lw=.5, alpha=.4, ymin=0, ymax=.1):
    """Draw a whisker at the bottom of the plot for each value of x

    See also plt.eventplot()

    This is useful if your points are clustered together too closely to be seen, but for 
    some reason the densityPlot() isn't suitable for your plot.

    This implementation uses a LineCollection class to speed rendering.
    Inputs
    ---------
    x
        (1d np array) List of x values to plot
    
    Optional Inputs
    -------------------
    clr
        (str) Colour of lines, in a format matplotlib expects 
    lw
        (float) Width of lines
    alpha
        (float) transparency
    ymin
        (float) Bottom of whisker in axis units 
    ymax
        (float) Top of whisker in axis units.

    Returns
    ---------
    **None**
    """
    #Matplotlib expects the data in this odd format. [row, line_elt, xOrY]
    data = np.empty( (len(x), 2, 2) )
    data[:,0,0] = x
    data[:,0,1] = 0
    data[:,1,0] = x
    data[:,1,1] = .1 

    ax = plt.gca()
    trans = ax.get_xaxis_transform(which='grid')
    collection = mcollect.LineCollection(data, linewidths=lw, colors=clr, alpha=alpha, transform=trans)
    ax.add_collection(collection)


def borderplot(x, y, *args, **kwargs):
    """Plot x against y, and show histograms in x and y at the same time.

    Inputs
    ---------
    x, y
        (float) Data to plot

    Optional Inputs
    ----------------
    xbins, ybins
        (int, array, str) Binning specification for the histogram
        in the x and y directions. Is passed directly to `plt.hist`
    color
        (matplotlib colour specification
    func
        (function) What function is used to draw main panel. Default is plt.plot

    All other options are passed to `plt.plot`

    Returns
    ---------
    ax
        Axis handle for main plot
    xh
        Axis handle for x-axis histogram
    yh
        Axis handle for y-axis histogram
    """

    xbins = kwargs.pop('xbins', 10)
    ybins = kwargs.pop('ybins', 10)
    color = kwargs.pop('color', 'C0')
    func = kwargs.pop('func', plt.plot)

    #Set up subplots
    i0, j0 = 2, 7
    gs = GridSpec(9,9)
    ax = plt.subplot(gs[i0:, :j0])
    func(x, y, *args, **kwargs)

    xh = plt.subplot(gs[:i0, :j0], sharex=ax)
    plt.hist(x, bins=xbins, color=color)
    xh.xaxis.tick_top()

    yh = plt.subplot(gs[i0:, j0:], sharey=ax)
    plt.hist(y, bins=ybins, color=color, orientation='horizontal')
    yh.yaxis.tick_right()

    plt.sca(ax)
    return ax, xh, yh




def densityPlot(x,y, xBins, yBins, *args, **kwargs):
    """
    Plot x against y as points where points are sparse, but as shaded region when crowded.

    When you have a lot of points to plot in a 2d graph, it is often difficult
    to decide the best way to plot it. If you are interested in the typical
    behaviour, an image of a 2d histogram is best, showing the density of
    points in different regions. If you're interested in the outliers it
    is better to plot each point individually.

    ``densityPlot`` combines the best of both worlds, showing indivdual
    outliers overlaid on a density plot of a 2d histogram.


    Inputs:
    -------
    x, y
        (np.array) Arrays of values to plot
    xbins, yBins
        What bins to use for density 2d histogram. Can be ints or arrays
        as allowd by ``np.histogram2d``

    Optional Inputs:
    ---------------
    threshold
        Individual data points are plotted if there are fewer than this
        number in a 2d histogram bin.
    cmap
        Which colour map to use for histogram. Default: mp.cm.rainbow
    aspect
        Default value scales axes to give square bins.

    All other optional arguments are passed to ``mp.plot()``


    Returns:
    --------
    **None**

    Notes:
    -----
    * Inspired by a similar IDL function written by Micheal Strauss in IDL
    * Won't work with postscript output
    * Best results are obtained by using a colour map similar to the
      colour of the points your using, e.g mp.cm.Greens and green circles.

    TODO:
    ---------
    Instead of having a wantLog, I should be able to pass in a normalisation
    object.
    """
    threshold = kwargs.pop('threshold', 10)
    vmin = kwargs.pop('vmin', None)
    vmax = kwargs.pop('vmax', None)
    norm = kwargs.pop('norm', fnorm.DiscreteNorm(7, vmin, vmax))
    cmap = kwargs.pop('cmap', DEFAULT_CMAP)

    ls = kwargs.pop('ls', "none")
    ls = kwargs.pop('linestyle', ls)
    marker = kwargs.pop('marker', '.')
    # aspect = kwargs.pop('aspect', 'auto')   #Equal gives square bins
    zorder = kwargs.pop('zorder', 0)

    cmap = copy.copy(cmap)  #I'm annoyed with mpl for making me do this
    cmap.set_under(alpha=0)
    plt.plot(x, y, zorder=zorder-1, marker=marker, ls=ls, *args, **kwargs)
    plt.hist2d(x, y, bins=[xBins, yBins], cmap=cmap, norm=norm)
    plt.clim(vmin=threshold)




def fix_date_labels():
    """Format date strings in x-axis label so they're easier to read
    
    Note
    ----------
    To get minor ticks once per day add the following code
    to your function after you call fix_date_labels()::

        import matplotlib.ticker as mticker
        ax.xaxis.set_minor_locator(mticker.IndexLocator(1, 0))

    This only works if you want one tick per day, and isn't helpful
    for much longer or shorter date spans

    """
    plt.gcf().autofmt_xdate()

    #If the xaxis range is of order 1 month, set the major ticks
    #to be once weekly
    locator = mdates.AutoDateLocator(minticks=2, maxticks=7, interval_multiples=False)
    locator.intervald['DAILY'] = 7
    ax = plt.gca()
    ax.xaxis.set_major_locator(locator)



# def load_figfile(filename):
#     """Load a figure from a pickle, similar to Matlab's .fig format"""
#     fig = plt.gcf()
#     with open(filename, 'rb') as fp:
#         fig = pickle.load(fp)
#     return fig

def load_figfile(filename):
    """Load a figure from a pickle, similar to Matlab's .fig format"""
    fig = plt.gcf()
    with open(filename, 'rb') as fp:
        bytes = fp.read()
        try:
            bytes = gzip.decompress(bytes)
        except gzip.BadGzipFile:
            # Older versions of save_figfile did not compress their data
            pass

        fig = pickle.loads(bytes)
    return fig


def mark_indices(x_data, y_axis, *args, **kwargs):
    """Place points with x-position referenced to data, but y-coords
    references to axis coords. This provides a kind of heads-up-display
    for when a binary event occurs (e.g a flag being raised on bad data)
    """
    if not hasattr(y_axis, 'len'):
        y_axis = y_axis + np.zeros_like(x_data, dtype=float)
        
    ax = plt.gca()
    trans = mtrans.blended_transform_factory(ax.transData, ax.transAxes)
    return ax.plot(x_data, y_axis, *args, transform=trans, **kwargs)



def mark_weekends(timestamps, tz='UTC'):
    """Plot grey bars to indicate night time in the central timezone"""
    t1 = min(timestamps)
    t2 = max(timestamps)

    t1 = "%04i-%02i-%02i 00:00" %(t1.year, t1.month, t1.day)  #10pm
    t2 = "%04i-%02i-%02i 00:00" %(t2.year, t2.month, t2.day)  #10pm
    # print (t1, t2)

    day_start = pd.date_range(start=t1, end=t2, freq='D')
    day_start = day_start.tz_localize(tz)

    #Edge case: First day of data set is Sunday
    day = day_start[0]
    if day.dayofweek == 6:
        delta = pd.to_timedelta("1D")
        handle = plt.axvspan(day-delta, day+delta, color='b', alpha=.1)
        plt.xlim(xmin=min(timestamps))

    for day in day_start:
        if day.dayofweek == 5:
            delta1 = pd.to_timedelta("6H")
            delta2 = pd.to_timedelta("30H")
            handle = plt.axvspan(day-delta1, day+delta2, color='b', alpha=.1)

    try:
        handle.set_label("Weekend")
    except UnboundLocalError:
        #No weekends marked
        pass


def mark_nights(timestamps, tz='UTC'):
    """Plot grey bars to indicate night time in the central timezone"""
    t1 = min(timestamps)
    t2 = max(timestamps)

    t1 = "%04i-%02i-%02i 22:00" %(t1.year, t1.month, t1.day)  #10pm
    t2 = "%04i-%02i-%02i 22:00" %(t2.year, t2.month, t2.day)  #8am
    # print (t1, t2)

    day_start = pd.date_range(start=t1, end=t2, freq='D')
    day_start = day_start.tz_localize(tz)

    delta = pd.to_timedelta("10H")
    for day in day_start:
        handle = plt.axvspan(day, day+delta, color='b', alpha=.1)

    #Edge case: First day of data set is Sunday
    # day = day_start[0]
    # if day.dayofweek == 6:
    #     delta = pd.to_timedelta("1D")
    #     handle = plt.axvspan(day-delta, day+delta, color='b', alpha=.1)
    #     plt.xlim(xmin=min(timestamps))

    # for day in day_start:
    #     if day.dayofweek == 5:
    #         delta1 = pd.to_timedelta("6H")
    #         delta2 = pd.to_timedelta("30H")
    #         handle = plt.axvspan(day-delta1, day+delta2, color='b', alpha=.1)

    try:
        handle.set_label("10pm-8am")
    except UnboundLocalError:
        #No weekends marked
        pass


def outline(clr='k', lw=1):
    """Add an outline to text or a line.
    
    See also shadow()
    Usage:
    ----------
    `plt.text(x, y, text, ... path_effects=fplots.outline() )`

    See https://matplotlib.org/stable/tutorials/advanced/patheffects_guide.html
    """

    return [
        meffect.Stroke(linewidth=lw, foreground=clr),
        meffect.Normal()
    ]



def plot_model_hist(model, bins, n_elt, *args, **kwargs):
    """Plot a stats model scaled so model.cdf(np.inf) == n_elt
    
    The scipy.stats module contains a number of common statistical functions
    (like normal, gumbel, laplacian, etc.) with a consistent interface. I often
    have the problem that I fit one of these models to some data in a maximum
    likelihood sense, then want to plot the model of a historgram of the data.
    The problem is that the model returns normalised values, i.e model.cdf(np.inf) = 1
    regardless of how many data points were including in the fit. 

    I could normalise the histogram of my data points, but that hides useful information.
    Instead, this function scales the values returned by model so they match the histogram.

    Usage:
    ---------
    ```
    import scipy.stats as spstats
    
    x = np.rand.randn(1000)
    pars = spstats.norm.fit(x)
    obj = spstats.norm(*pars)
    _, bins, _ = plt.hist(x, bins='fd')
    plot_model_hist(model, bins, len(x))
    ```

    Inputs
    -----------
    model
        (scipy.stats.rvs_continuous object) A frozen stats model (i.e one where the loc and 
        scale params have been set)
    bins
        (1d np array) location of bin edges of histogram
    n_elt
        (int) Number of elements used to create histogram of data from where the model was generated

    All other arguments are passed to matplotlib's step function

    Returns
    ----------
    **None**

    Output
    ----------
    A step function is overlaid the current plot
    """
    kwargs['where'] = kwargs.pop('where', 'post')
    vals = np.diff(model.cdf(bins))
    plt.step(bins[:-1], vals * n_elt, *args, **kwargs)


def monthly_plot(df, *args, plotter=None, datecol='date', ycol='val', **kwargs):
    """Plot a 4x3 grid of time-series, one per calendar month

    Useful to track seasonal changes in daily time-series

    Inputs
    --------
    df
        (DataFrame)
    plotter
        (Callable) A function to plot a month's worth of data. See below
    datecol
        (str) Name of column in `df` that contains the data information
    ycol
        (str) Name of column in `df` containing values to plot

    All other arguments are passed to the `plotter` function

    Returns
    ----------
    A dictionary of axis objects

    Notes
    -----
    * If no value passed for `plotter`, then `default_monthly_plotter` is called.
    This function serves as the reference implmentation.

    * The signture to the `plotter` function is
        * df (the dataframe)
        * datecol
        * args and kwargs.
    If ycol is set in the call to the parent function, it is passed as kwargs.
    """


    plotter = plotter or default_monthly_plotter
    kwargs['ycol'] = ycol

    df[datecol] = pd.to_datetime(df[datecol])
    df = df.sort_values(datecol)
    year = df[datecol].dt.year
    month = df[datecol].dt.month

    plt.clf()
    # grid = plt.gcf().subplot_mosaic("abc\ndef\nghi\njkl", sharex=True, sharey=True)
    grid = plt.gcf().subplot_mosaic("abcd\nefgh\nijkl", sharex=True, sharey=True)
    kwargs['_plot_grid'] = grid
    grid = df.groupby([year, month]).apply(_plot_single_month, plotter, datecol, *args, **kwargs)

    plt.subplots_adjust(wspace=0)
    return grid.iloc[0]  #Is a dictionary

def _plot_single_month(df, plotter, datecol,  *args, **kwargs):
    """Private function of `monthly_plot`"""
    #Pick the correct subplot
    subplots = '_abcdefghijkl'
    grid = kwargs.pop('_plot_grid')
    month = df[datecol].dt.month.iloc[0]
    plt.sca(grid[ subplots[month]])

    plotter(df, datecol, *args, **kwargs)
    months = "___ Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec".split()
    plt.title(months[month])
    return grid


def default_monthly_plotter(df, datecol, *args, **kwargs):
    """Example plotter for `monthly_plot()`
    
    Is passed a dataframe where every data point shares the same year and month.
    Any return value is ignored.
    
    Inputs
    -------
    df
        (DataFrame)
    datecol
        (str) Name of column to plot on the xaxis 
    
    All other arguments are passed to the `plt.plot` function
    """
    def plot(df, datecol, *args, **kwargs):
        ycol = kwargs.pop('ycol')
        log = kwargs.pop('logy', False)

        dates = df[datecol]
        frac_hour = dates.dt.hour + dates.dt.minute / 60
        assert np.all(np.diff(frac_hour) >= 0)
        plt.plot(frac_hour, df[ycol], *args, **kwargs)

        if log:
            plt.gca().set_yscale('log')

    day = df[datecol].dt.day
    df.groupby(day).apply(plot, datecol, *args, **kwargs)


def plot_with_discrete_cb(func, *args, **kwargs):
    """Create a plot with a discrete colorbar.

    Orbital Insight plot style encourages colourbars with 7 discrete
    colours instead of continous colors. This function wraps the call
    to a plotting function with one that creates such a discrete colourbar.

    Example plots shown in
    https://orbitalinsight.atlassian.net/wiki/spaces/DATSCI/pages/83394804/Fonts+and+Plotting+Themes_


    Inputs
    -----------
    func : Function to be wrapped. All required and optional arguments should be passed to
        ``plot_with_discrete_cb`` as they would the original function.

    Optional Inputs
    ----------------
    nstep (int): How many steps in the discrete colormap
    cmap (``matplotlib.colors.Colormap`` object): The colourmap to use
    clim (2 element list of floats):  Max and min values to map to the colour map.
        This allows you to control what range of values get pinned to the extremes
        of the colour range
    cb_ticklabel_format (Format string): Control the format of the tick labels for the colourbar.

    All other optional inputs get passed to ``func``


    Example
    ------------
    .. code-block:: python
        plot_with_discreate_cb(plt.scatter, x, y, s=z, c=z, cmap=plt.cm.Greens)

    """

    nstep = kwargs.pop('nstep', 8)
    cmap = kwargs.pop('cmap', plt.cm.Greens)
    tick_format = kwargs.pop('cb_ticklabel_format', mticker.FormatStrFormatter("%i"))
    clim = kwargs.pop('clim', None)

    if clim is None:
        raise ValueError("clim must be specified")

    norm = mcolors.BoundaryNorm(np.linspace(clim[0], clim[1], nstep), cmap.N)
    kwargs['norm'] = norm
    kwargs['cmap'] = cmap

    function_handle = func(*args, **kwargs)
    cb_handle = plt.colorbar(format=tick_format)
    return function_handle, cb_handle

from frmbase.plateau import plateau
def plot_with_gaps(x, y, *args, gap_size=None, **kwargs):
    """Plot a timeseries skipping over any gaps.
    Useful for linecharts where the x values are not regularly spaced.
    Inputs
    -------
    same as plt.plot except for the keyword argument gap_size
    If the difference between two values in x[] are larger than gapsize
    then no connector is drawn between them.

    The default gapsize is 1% the span of the data
    Note
    ------
    If the x value is in datetime format, gap_size should be be pd.to_timedelta()
    object
    """
    if gap_size is None:
        gap_size = .01 * (np.max(x) - np.min(x))

    diff = np.diff(x)
    #casting works around issues with dates
    # peaks = plateau(diff.astype(float) > gap_size, .5)
    peaks = plateau(diff > gap_size, .5)
    # import ipdb; ipdb.set_trace()

    peaks = np.concatenate([[int(0)], peaks.flatten(), [int(len(x))]]).reshape(-1, 2).astype(int)
    for p in peaks:
        lwr, upr = p
        plt.plot(x[lwr:upr], y[lwr:upr], *args, **kwargs)


def ternary(v1, v2, v3, labels=None, **kwargs):
    """Create a ternary plot

    https://en.wikipedia.org/wiki/Ternary_plot

    A ternay plot is basically a 2d representation of a plane in
    3 dimensions. For each element of the three input vector, the
    sum of `v1[i] + v2[i] + v3[i] == 1`

    Inputs
    --------
    v1, v2, v3 (1d arrays)
        Values to plot. v1[i] + v2[i] + v3[i] == 1 for all values of i,
        and 0 <= vj[i] <= 1 for all values of i,j
    labels (list of strings)
        Text to annotate the 3 corners of the triangle.

    Optional Arguments
    ----------------------
    decorate (bool)
        If **True**, some helpful contour lines are drawn on the plot

    All other optional arguments are passed to plt.scatter

    Notes
    --------
    The solid horizontal lines drawn by decorate indicate the strength
    of vector 1. The dashed lines indicate the relative strength
    of vector 3 over vector 2. This was useful for my first use case,
    and may not make sense in other use cases.

    """
    assert len(v1) == len(v2)
    assert len(v2) == len(v3)

    for v in [v1, v2, v3]:
        assert 0 <= np.min(v) and  np.max(v) <= 1
    assert np.allclose(v1 + v2 +v3, 1)

    if labels is None:
        labels = ["Vector 1", "Vector 2", "Vector 3"]
    assert len(labels) == 3

    want_decoration = kwargs.pop('decorate', False)

    #Draw and label the triangle
    x = [-.5, 0, .5, -.5]
    y = [0, np.sqrt(3)/2., 0, 0]
    plt.plot(x, y, 'k-', lw=2, zorder=-1)

    plt.text(0, .866, "  " + labels[0], ha="left", fontsize=18)
    plt.text(.5, -.05, labels[1], ha="center", fontsize=18)
    plt.text(-.5,-.05, labels[2], ha="center", fontsize=18)

    #Mark the points
    coords = _compute_ternary_coords(v1, v2, v3)
    plt.scatter(coords[:, 0], coords[:, 1], **kwargs)

    #Scale and hide the axis.
    plt.axis('equal')
    plt.axis('off')

    if want_decoration:
        _decorate_ternary_plot(labels)


def _compute_ternary_coords(a, b, c):
    a = np.array(a)
    b = np.array(b)
    c = np.array(c)
    # #Convert pandas series to numpy arrays if necessary
    # try:
    #     a = a.values
    #     b = b.values
    #     c = c.values
    # except AttributeError:
    #     pass

    root3 = np.sqrt(3)
    y = root3 * a / 2.
    x = .5 * (b-c)
    return np.vstack([x,y]).transpose()

def _decorate_ternary_plot(labels):
    #Draw horizontal lines
    for zz in [.25, .5, .75]:
        zy = (np.sqrt(3)/2 * (zz))
        zx = .5*(1-zz)
        print(zz, zx, zy)
        plt.text(zx, zy, " %.0f%% %s" %(100*zz, labels[0]))

        plt.plot([-zx, zx], [zy, zy], '-', color='grey', lw=.5)
        plt.xlim(xmax=.6)

    #Draw diagonal lines
    # for zz in [.25, .5, .75]:
    #     zx = -.5 + zz
    #     zy = 0
    #     # plt.plot([zx, zx], [zy, zy+.01], 'k-')
    #     plt.plot([0, zx], [.866, 0], 'k-', lw=1, ls="--", zorder=-2)
    #     text = "%i%% %s" %(100*zz, labels[1])
    #     plt.text(zx, -0.04, text, fontsize=10, ha='center')

    p0 = _compute_ternary_coords([0, .75], [.25, .25], [.75, 0])
    plt.plot(p0[:,0], p0[:,1], 'b-', lw=1, ls="--", zorder=-2)
    text = "%i%% %s " %(25, labels[1])
    plt.text(p0[0,0], p0[0,1]-.2, text, rotation=60, va='bottom', ha='right')

    p0 = _compute_ternary_coords([0, .5], [.5, .5], [.5, 0])
    plt.plot(p0[:,0], p0[:,1], 'b-', lw=1, ls="--", zorder=-2)
    text = "%i%% %s " %(50, labels[1])
    plt.text(p0[0,0], p0[0,1]-.2, text, rotation=60, va='bottom', ha='right')


# def save_figfile(filename):
#     """Save a figure in a pickle, similar to Matlab's .fig format"""
#     fig = plt.gcf()
#     with open(filename, 'wb') as fp:
#         pickle.dump(fig, fp)

def save_figfile(filename):
    """Save a figure in a compressed pickle, similar to Matlab's .fig format"""
    fig = plt.gcf()
    with open(filename, 'wb') as fp:
        fp.write( gzip.compress(pickle.dumps(fig)) )


def shadow(**kwargs):
    """Add a drop-shadow to a line or some text 

    See also outline()
    
    Usage:
    ----------
    `plt.plot(x, y, ... path_effects=fplots.shadow() )`

    See https://matplotlib.org/stable/tutorials/advanced/patheffects_guide.html
    """

    return [meffect.SimpleLineShadow(**kwargs), meffect.Normal()]


def text_at_axis_coords(x, y, text, *args, **kwargs):
    """Place text at coords relative to axis. 0,0 always means bottom left of plot"""

    ax = plt.gca()
    ax.text(x, y, text, *args, transform=ax.transAxes, **kwargs)


def text_at_figure_coords(x, y, text, *args, **kwargs):
    """Place text at coords relative to the figure. 0,0 always means bottom left of figure"""

    ax = plt.gca()
    f = plt.gcf()
    ax.text(x, y, text, *args, transform=f.transFigure, **kwargs)






from  ipdb import set_trace as idebug 
from matplotlib.widgets import Cursor
import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import matplotlib as mpl
import pandas as pd
import numpy as np


class AbstractInteractivePlot():
    def __init__(self):
        #Set input arguments as class variables.

        self.fig1 = plt.gcf()
        self.fig2 = plt.figure()

        self.updateEvent = 'key_press_event'
        self.fig1.canvas.mpl_disconnect(self.updateEvent)
        self.fig1.canvas.mpl_connect(self.updateEvent, self)

        plt.figure(self.fig1.number)
        plt.clf()
        self.plotFunc()
        plt.pause(.01)


    def __del__(self):
        self.disconnect()

    def __call__(self, event):
        print ("call")
        if not event.inaxes:
            return

        x = event.xdata
        y = event.ydata
        key = event.key
        print(x, y, key)

        if key == 'q':
            return 

        plt.figure(self.fig2.number)
        plt.cla()
        self.callback(x, y, key)
        plt.pause(.01)
        plt.figure(self.fig1.number)

    def disconnect(self):
        self.fig1.canvas.mpl_disconnect(self.updateEvent)

    def plotFunc(self,):
        raise NotImplementedError

    def callback(self, x, y, key):
        raise NotImplementedError


class AbstractDataframeInteractivePlot(AbstractInteractivePlot):
    def __init__(self, df, xCol, yCol):
        self.df = df 
        self.xCol = xCol 
        self.yCol = yCol 

        AbstractInteractivePlot.__init__(self) 

    def xyToRow(self, x, y):
        """Utility function to convert x,y to a row in a dataframe"""
        xvals = self.df[self.xCol].values
        yvals = self.df[self.yCol].values

        #Work around date issues
        if isinstance(xvals[0], pd.Timestamp):
            print("Is datetime")
            xvals = mdates.date2num(xvals)
        dx = xvals - x
        dy = yvals - y

        #Convert abs distance in each direction to distances relative to axis limits
        #this behaves better when the x and y axes cover dramatically different ranges
        x1, x2, y1, y2 = plt.axis()
        frac_x = dx / (x2 - x1)
        frac_y = dy / (y2 - y1) 

        dist = np.hypot(frac_x, frac_y)
        dist[ ~np.isfinite(dist) ] = 2 * np.max(dist)
        row = np.argmin(dist)
        return row

class InteractivePlot(AbstractDataframeInteractivePlot):
    """
    
    This is an example implementation

    Clicking in the plot activates the callback function which
    prints the relevant row in the dataframe

    See exampleInteractivePlot() in this file
    """
    def __init__(self, df, xCol, yCol):
        """

        Inputs:
        -------
        data
            (np.ndarray) Data to plot an interact with

        xCol, yCol
            (ints) Which columns of x and y to plot. See note below



        Notes:
        ---------
        plotFunc will be passed xCol and yCol, but is not required to
        do anything with them. However, when the plot is clicked,
        the object computes the row that minimises
        ::
          hypot( data[row, xCol] - xClick, data[row, yCol] - yClick)

        and passed that row to the call back function. It will be
        pretty confusing to the user if the xcol and ycol of data
        are not plotted on the screen when this happens.

        """

        AbstractDataframeInteractivePlot.__init__(self, df, xCol, yCol)

    def plotFunc(self,):
        xCol, yCol = self.xCol, self.yCol
        plt.clf()
        plt.plot(self.df[xCol], self.df[yCol], 'C0o')

    def callback(self, x, y, key):
        row = self.xyToRow(x, y)
        print( key, self.df.iloc[row])





class Crosshairs:
    """
    A cross hair cursor.

    Draw a curson consisting of vertical and horizontal lines centered
    on the mouse. The cursor updates with each mouse move.

    """
    def __init__(self, ax=None):
        if ax is None:
            ax = plt.gca()

        self.ax = ax
        self.horizontal_line = ax.axhline(color='k', lw=0.8, ls='--')
        self.vertical_line = ax.axvline(color='k', lw=0.8, ls='--')
        # text location in axes coordinates
        self.text = ax.text(0.72, 0.9, '', transform=ax.transAxes)
    
        plt.gcf().canvas.mpl_connect('motion_notify_event', self.on_mouse_move)

    def set_cross_hair_visible(self, visible):
        need_redraw = self.horizontal_line.get_visible() != visible
        self.horizontal_line.set_visible(visible)
        self.vertical_line.set_visible(visible)
        self.text.set_visible(visible)
        return need_redraw

    def on_mouse_move(self, event):
        if not event.inaxes:
            need_redraw = self.set_cross_hair_visible(False)
            if need_redraw:
                self.ax.figure.canvas.draw()
        else:
            self.set_cross_hair_visible(True)
            x, y = event.xdata, event.ydata
            # update the line positions
            self.horizontal_line.set_ydata(y)
            self.vertical_line.set_xdata(x)
            self.text.set_text('x=%1.2f, y=%1.2f' % (x, y))
            self.ax.figure.canvas.draw()



# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import division

import matplotlib.pyplot as mp
import matplotlib as mpl
import numpy as np

import datetime as dt
import inspect
import os

"""
Site-customisation 

Imports the Orbital Insight style file, and provides functions for
annotating a plot. Functions include

add_minor_ticks()
    Add minor ticks to plot axes, which matplotlib does not do by default
    unless the plot is log scaled.

add_watermark()
    Add a small identifying string to the plot indicating the plot
    creator, date, and source file

mark_as_draft()
    Write Draft is big grey letters over the plot

"""

#Windows specific
if 'USER' not in os.environ:
    os.environ['USER'] = 'fergal'

#Auto import the style file. Is this a good idea?
try:
    fname = os.path.join(mpl.get_configdir(), 'fergal.mplstyle')
    mpl.style.use(fname)
except IOError:
    print ("WARN: Fergal's style file not found")
    print ("Config dir is %s" %(mpl.get_configdir()))


def add_minor_ticks():
    """Add minor ticks to axes"""
    ax = mp.gca()
    for a in [ax.xaxis, ax.yaxis]:
        if a.get_scale() == 'linear':
            a.set_minor_locator(mpl.ticker.AutoMinorLocator())

        a.set_tick_params(which="major", length=8, width=1, color='k')
        a.set_tick_params(which="minor", length=4, width=1, color='k')


def remove_minor_ticks():
    ax = mp.gca()
    for a in [ax.xaxis, ax.yaxis]:
        a.set_minor_locator(mpl.ticker.NullLocator())


def add_watermark(level=0, loc='right'):
    """Put an mark on the bottom right of a figure identifing
    plot creator, date, and source file.

    Optional Inputs
    -----------------
    level
        (int) How many levels up the calling stack is the function whose
        name should be included in the watermark test. Default is the function
        that calls add_watermark()

        Example::

            foo():
                bar():
                    add_watermark(0)  #Watermark says $USER code.py:bar ...
                    add_watermark(1)  #Watermark says $USER code.py:foo ...
    """

    text = create_watermark_text(level+1)
    ax = mp.gca()
    f = mp.gcf()

    if loc == 'right':
        ax.text(0.93, 0.88, text, rotation=90, size=8, transform=f.transFigure, ha="right", va="top")
    elif loc == 'bottom':
        ax.text(0.95, 0.01, text, size=8, transform=f.transFigure, ha="right")
    elif loc == 'top':
        ax.text(0.95, 0.96, text, size=8, transform=f.transFigure, ha="right")
    else:
        raise ValueError("loc should be one of 'right', 'top', or 'bottom'")

    
def create_watermark_text(level=0):
    """Create the text string for the watermark. See add_watermark for details"""
    user = os.environ['USER']

    full_fn = '/.'
    try:
        frame = inspect.currentframe().f_back
        for i in range(level):
            try:
                frame = frame.f_back
            except AttributeError:
                raise AttributeError("Requested level (%i) exceeds stack depth (%i)" %(level, i))

        full_fn = frame.f_globals['__file__']
        fn = os.path.basename(full_fn)

        func = frame.f_code.co_name
        fn = "%s:%s" %(fn, func)
    except KeyError:
        fn = "[Shell]"
    time = dt.datetime.strftime( dt.datetime.now(), '%Y-%m-%d %H:%M')

    try:
        import frmbase.meta as meta
        git_meta = meta.get_git_info(full_fn)
        git_hash = git_meta['__git_branch_commit_sha__']
        git_status = git_meta['__git_commit_status__'][0] #One char
        git_hash = "%s (%s)" %(git_hash, git_status)
    except (ImportError, KeyError):
        git_hash = ""

    text = "%s %s %s %s" % (user, fn, time, git_hash)
    return text 


def mark_as_draft(text="Draft"):
    """Write the word draft on a plot in big letters

    Optional Inputs:
    --------------
    text
        Text to write

    """
    ax = mp.gca()
    mp.text(.5, .5, text, va="center", ha="center", fontsize=80, color='k',
            alpha=.4, transform=ax.transAxes, rotation=45)




from matplotlib.gridspec import GridSpec
import matplotlib.pyplot as plt
import numpy as np


def galaxyPlot(df, colList, *args, **kwargs):
    """Plot everything against everything

    Just like extra galactic astronomers are wont to do.

    Inputs:
    -------------
    df
        (Dataframe) Data to be plotted

    colList
        (list) list of column numbers to plot. Must have at least
        two elements

    Optional Inputs:
    ---------------
    plotFunc
        (A plotting function) What function should actually plot the x,y
        points. Default is plt.plot()

    labels
        (list of strings) Applied as x and y axis labels. If not
        given, defaults to values in colList

    wantLog
        (list or array) If True, given column is plotted in log scale.
        len(wantLog) == len(colList)

    limits
        (list) If the ith elememt of this array is an object of length 2,
        use those two number to set the plot limits in that dimension.
        See notes, below.

    All other optional arguments are passed to plt.plot (or whatever plotFunc
    is)

    Returns:
    -------
    void

    Output:
    --------
    A plot is produced.


    Notes:
    -------
    The signature of plotFunc is
    ``plotFunc(x, y, *args, **kwargs)``
    similar to plt.plot()'s most commonly called style.

    The easiest way to create the limits argument is as follows::

      limits = list(np.zeros_like(colList))
      limits[4] = [1, 10]

    The function will see that only the 4th element of the array
    is a set of limits, and default values will be chosen for all
    other dimensions

    TODO
    ------
    * Adapt code to convert input numpy arrays to dataframes.
    * colList should be an optional argument

    """

    labels = kwargs.pop('labels', colList)
    ms = kwargs.pop('markersize', 3)
    plotFunc = kwargs.pop('plotFunc', plt.plot)

    wantLog = np.zeros(len(colList))
    wantLog = kwargs.pop('wantLog', wantLog)
    assert( len(wantLog) == len(colList))

    limits = np.zeros(len(colList), dtype=bool)
    limits = kwargs.pop('limits', limits)
    assert( len(limits) == len(colList))

    #TODO Convert numpy to dataframe if numpy is input

    nCol = len(colList)
    nPlot = nCol-1

    #Create a 2d array of Axes objects
    axArray = np.empty( (nCol, nCol), dtype=object)
    for i in range(nCol):
        for j in range(0, i):
            n = (nPlot)*(i-1)+j + 1
            sharex, sharey = findSharedAxes(axArray, i, j)

            if sharex is None and sharey is None:
                ax = plt.subplot(nPlot, nPlot, n)
            if sharex is None and sharey is not None:
                ax = plt.subplot(nPlot, nPlot, n, sharey=sharey)
            if sharex is not None and sharey is None:
                ax = plt.subplot(nPlot, nPlot, n, sharex=sharex)
            if sharex is not None and sharey is not None:
                ax = plt.subplot(nPlot, nPlot, n, sharex=sharex, sharey=sharey)

            axArray[i,j] = ax

    #i loops over the y-axes (the rows)
    for i in range(nCol):

        #Pick y limits for subplot. If not supplied, use range of data
        if hasattr(limits[i], "__len__") and len(limits[i]) == 2:
            yLwr, yUpr = limits[i]
        else:
            yLwr = np.nanmin(df[colList[i]])
            yUpr = np.nanmax(df[colList[i]])

        #j loops over the x-axes (the columns)
        for j in range(0, i):
            ax = axArray[i,j]
            plt.sca(ax)

            plotFunc(df[colList[j]], df[colList[i]], \
                markersize=ms, *args, **kwargs)

            #Add decorations for the subplot
            if j== i-1:
                plt.title(labels[j])

            if j == 0:
                plt.ylabel(labels[i])

            if i != nCol-1:
                plt.setp(ax.get_xticklabels(), visible=False)
            else:
                plt.xlabel(labels[j])

            if wantLog[i]:
                ax.set_yscale('log')
                ax.yaxis.set_tick_params(which="major", length=10)
                ax.yaxis.set_tick_params(which="minor", length=4)


            if wantLog[j]:
                ax.set_xscale('log')
                ax.xaxis.set_tick_params(which="major", length=10)
                ax.xaxis.set_tick_params(which="minor", length=4)

            if j > 0:
                plt.setp(ax.get_yticklabels(), visible=False)

            #Pick x limits for subplot. If not supplied, use range
            #range of data
            if hasattr(limits[j], "__len__") and len(limits[j]) == 2:
                plt.xlim(limits[j])
            else:
                xLwr = np.min(df[colList[j]])
                xUpr = np.max(df[colList[j]])
                plt.xlim([xLwr, xUpr])
            plt.ylim([yLwr, yUpr])

    plt.subplots_adjust(wspace=0, hspace=0)


def findSharedAxes(axArray, i, j):
    xAx = None
    yAx = None

    if i>j+1:
        xAx = axArray[i-1, j]

    if j>0:
        yAx = axArray[i, j-1]

    return xAx, yAx



# -*- coding: utf-8 -*-

"""
Some matplotlib normalization objects with discrete colourbars.

Mapping values to colours in matplotlib (e.g for scatter plots or heatmaps)
is a two step process. First the values are mapped to the range [0,1], then
the normalised values are mapped to individual colours. The first step
is performed by a normalisation object, the second by a colour map.

Matplotlib comes with two useful normalization objects, a linear and a
logarithm norm. These map the input values smoothly between zero and one.
However, for many applications, it is more visually useful to map input
values to a small number of discrete ranges (e.g. zero to twenty all gets
mapped to a single colour, twenty to fourty to another colour, etc.)

The normalization objects in this module provide two options for discrete
normalziation; Linear, and Histogram Equalisation. Linear creates a mapping
where the range of values covered by each colour is equal. Histogram
Equalisation chooses the ranges so the same number of data points fall
within in each bin.
"""

from __future__ import print_function
from __future__ import division

import matplotlib.colors as mcolors
import numpy as np

eps = np.finfo(np.float32).eps

class DiscreteNorm(mcolors.BoundaryNorm):
    """Scale a collection of numbers to lie in the range [0,1]
    so that the normalized values can only take `ncolors` discrete values

    This is useful for creating colorbars that are segmented into a
    small number of discrete colours. Small numbers of colours (<8)
    are easier to distinguish than the default hundrets of colours.

    This class maps the input to output range in a linear fashion.
    More complicated approaches are also possible

    Example
    ---------
    ::

        y = np.random.randn(1000)
        ncolors = 5  #Show 7 discrete colors in colorbar
        norm = DiscreteNorm(ncolors)
        plt.scatter(x, y, norm=norm)


    """

    def __init__(self, ncolors, vmin=None, vmax=None, clip=False):
        mcolors.Normalize.__init__(self, vmin, vmax, clip)
 
        # self._clip = clip
        self.ncolors = ncolors
        # self._vmin = vmin
        # self._vmax = vmax
        self.bounds = None

    @property
    def boundaries(self):
        if self.bounds is None:
            if self._vmin is None or self._vmax is None:
                raise ValueError("Unusual error in norm. Please set vmin and vmax")
            self.bounds = self.computeBoundaries()
        return self.bounds


    def autoscale_None(self, A):
        """autoscale only None-valued vmin or vmax."""
        # import ipdb; ipdb.set_trace()
        A = np.asanyarray(A)
        if self._vmin is None and A.size:
            self._vmin = A.min()
        if self._vmax is None and A.size:
            self._vmax = A.max()

        if self.bounds is None:
            self.bounds = self.computeBoundaries(A)


    def computeBoundaries(self, x=None):
        return np.linspace(self._vmin, self._vmax, self.ncolors+1)

    def __call__(self, values, clip=None):
        if clip is None:
            clip = self._clip

        xx, is_scalar = self.process_value(values)

        #Perform clipping if necessary
        if clip:
            np.clip(xx, self._vmin, self._vmax, out=xx)

        self.autoscale_None(xx)

        #Set norm values
        scale = float(len(self.bounds) -2 )
        iret = np.zeros(xx.shape, dtype=float)
        for i, b in enumerate(self.bounds):
            iret[xx > b] =  (i / scale)
        iret *= .999999

        #Set norm values for out-of-bounds inputs
        iret[xx < self._vmin] = -1
        iret[xx > self._vmax] = 2

        #Convert to masked array?
        mask = np.ma.getmaskarray(xx)
        ret = np.ma.array(iret, mask=mask)

        if is_scalar:
            return ret[0]

        return ret

    def inverse(self, value):
        raise ValueError("Not invertable")
#   def __call__(self, value, clip=None):
#         """
#         Normalize *value* data in the ``[vmin, vmax]`` interval into the
#         ``[0.0, 1.0]`` interval and return it.

#         Parameters
#         ----------
#         value
#             Data to normalize.
#         clip : bool
#             If ``None``, defaults to ``self.clip`` (which defaults to
#             ``False``).

#         Notes
#         -----
#         If not already initialized, ``self.vmin`` and ``self.vmax`` are
#         initialized using ``self.autoscale_None(value)``.
#         """
#         if clip is None:
#             clip = self.clip

#         result, is_scalar = self.process_value(value)

#         self.autoscale_None(result)
#         # Convert at least to float, without losing precision.
#         (vmin,), _ = self.process_value(self.vmin)
#         (vmax,), _ = self.process_value(self.vmax)
#         if vmin == vmax:
#             result.fill(0)   # Or should it be all masked?  Or 0.5?
#         elif vmin > vmax:
#             raise ValueError("minvalue must be less than or equal to maxvalue")
#         else:
#             if clip:
#                 mask = np.ma.getmask(result)
#                 result = np.ma.array(np.clip(result.filled(vmax), vmin, vmax),
#                                      mask=mask)
#             # ma division is very slow; we can take a shortcut
#             resdat = result.data
#             resdat -= vmin
#             resdat /= (vmax - vmin)
#             result = np.ma.array(resdat, mask=result.mask, copy=False)
#         if is_scalar:
#             result = result[0]
#         return result

class FixedDiscreteNorm(DiscreteNorm):
    """Set the boundaries between the colours directly"""
    def __init__(self, boundaries, vmin=None, vmax=None, clip=False):
        self.mclip = clip
        self.mvmin = vmin or np.min(boundaries)
        self.mvmax = vmax or np.max(boundaries)
        self.bounds = np.array(boundaries)
        self.ncolors = len(boundaries)

    def computeBoundaries(self, x=None):
        return self.bounds

        
class PercentileNorm(DiscreteNorm):
    """Not tested"""
    def __init__(self, levels, vmin=None, vmax=None, clip=False):
        DiscreteNorm.__init__(self, levels, vmin, vmax, clip)
        self.mclip = clip
        self.mvmin = vmin
        self.mvmax = vmax

        if isinstance(levels, int):
            levels = np.linspace(0, 100, levels+1)
        self.pct_levels= levels


    def computeBoundaries(self, x):
        eps = np.finfo(np.float32).eps
        x = np.array(x)  #Strip out array masking if necessary

        vmin = self.mvmin or np.min(x)
        vmax = self.mvmax or np.max(x)

        x = x[ (x >= vmin) & (x <= vmax)]

        boundaries = np.percentile(x, self.pct_levels)
        boundaries[-1] += eps
        boundaries = fixEqualValueBoundaries(boundaries)
        assert np.all(np.diff(boundaries) > 0), "Logic Error"
        return boundaries


class DiscreteLog10Norm(DiscreteNorm):
    def __init__(self, ncolors, vmin=eps, vmax=None, clip=False):
        DiscreteNorm.__init__(self, ncolors, vmin, vmax, clip)

    def computeBoundaries(self, x):
        x = np.array(x)
        x = x[ (x >= self._vmin) & (x <= self._vmax)]

        if np.any(x <=0):
            raise ValueError("Can't apply log normalisation to negative numbers")

        lwr = np.log10(np.min(x))
        upr = np.log10(np.max(x) + eps)
        boundaries = np.logspace(lwr, upr, self.ncolors+1)
        return boundaries


class HistEquNorm(DiscreteNorm):
    """Create a DiscreteNorm object with thresholds chosen so that each colour
    is equally represented in the figure. This can sometimes expose
    detail hidden by a linear colourmap.

    While DiscreteNorm() produces equally spaced thresholds between colors
    this object chooses the thresholds so that equal numbers of the input
    data points appear in each colour 'bin'. This is useful, for example,
    for datasets drawn from power law distributions.


    Example
    ---------
    ::

        y = np.random.randn(1000)
        ncolors = 5  #Show 7 discrete colors in colorbar
        norm = HistEquNorm(y, ncolors)
        plt.scatter(x, y, norm=norm)
        plt.colorbar(spacing='proportional'
    """

    def computeBoundaries(self, x):
        x = np.array(x)  #Strip out array masking if necessary
        thresholds = np.linspace(0, 100, self.ncolors+1)

        x = x[ (x >= self._vmin) & (x <= self._vmax)]
        boundaries = np.percentile(x, thresholds)
        boundaries[-1] += eps
        import ipdb; ipdb.set_trace()

        boundaries = fixEqualValueBoundaries(boundaries)
        assert np.all(np.diff(boundaries) > 0), "Logic Error"
        return boundaries


def fixEqualValueBoundaries(boundaries):
    """Adjust boundary values so no pair are equal

    For strange distributions of data it can happen that some of the
    boundaries chosen by percentiles (e.g in HistEquNorm.computeBoundaries)
    have equal values. This messes up the colorbar a bit.
    This function identifies when this issue occurs and adjusts
    the boundary values in nice ways to make the colourbar look
    prettier.

    It operates in two passes, a forward and reverse pass.
    The algorithm is quite slow in python (no vectoring) but
    it is typically run on arrays with fewer than 10 elements,
    so speed isn't much of an issue.

    Inputs
    ---------
    boundaries
        (list or array) Sorted list of boundary values

    Returns
    -----------
    Array of same length as input. Values in the output are
    adjusted to avoid duplicate values.
    """

    boundaries = _fixEqualValueBoundaries(boundaries)
    b = _fixEqualValueBoundaries(boundaries, reverse=True)
    return b


def _fixEqualValueBoundaries(boundaries, reverse=False):
    """See fixEqualValueBoundaries"""
    boundaries = np.array(boundaries).astype(float)

    #Check sorted
    assert np.all(np.diff(boundaries) >= 0)

    if reverse:
        boundaries = -1 * boundaries[::-1]

    diff = np.diff(boundaries)
    i = 0
    while i < len(diff):
        if diff[i] > 0:
            i += 1
            continue

        j = i+1
        while j < len(diff) and diff[j] == 0:
            j += 1

        if j == len(diff):
            #We've hit the end of the array, so stop
            return boundaries

        delta = boundaries[j+1] - boundaries[i]
        size = j - i
        boundaries[i:j+1] += np.arange(size+1) * delta / float(size+1)

        i = j

    if reverse:
        #Un-reverse
        boundaries = -1 * boundaries[::-1]
    return boundaries

#



"""

This will be a code to annotate a figure in such a way that labels don't overlap

It's not there yet
"""

from ipdb import set_trace as idebug
import matplotlib.pyplot as plt
import matplotlib as mpl
import numpy as np
import itertools

"""
if is_visible[j]:

    for h in [left, center, right]:
        t.set_horizontal_alignment(h)
        for v in ["baseline","bottom",  "top"]
            t.set_vertical_alignment(v)
            is_overlap = self.overlap(t, self.tList[j])
            if not is_overlap:
                #Break out of the loop, the label goes here"

"""


class AbstractPlatLabel():
    """Hide some overlapping labels on a plot

    Given a set of matplotlib.Text objects, figure out which
    ones overlap, and which ones should be hidden to avoid
    overlaps.

    Usage:
    ml= MapLabel()

    for i in range(10)
        handle = mp.text(i, i, "Some text")
        ml.add(handle)

    #At this point, all labels are shown, whether they overlap
    #or not. This next method call hides those that overlap
    ml.render()


    Note:
    ----------
    MapLabel objects can be combined with the + operator.
    mapLabel1 + mapLabel2 returns a new mapLabel object
    that contains handles to the text objects in both original
    objects. This can be convenient to collect text handles
    from functions before calling render once on the combined
    set.
    """

    def __init__(self):
        self.tList = []


    def __add__(self, mapLabel2):
        tList = []
        tList.extend(self.tList)

        try:
            tList.extend(mapLabel2.tList)
        except AttributeError:
            raise NotImplementedError()

        out = PlatLabel()
        out.tList = tList
        return out


    def __iadd__(self, mapLabel2):
        try:
            self.tList.extend( mapLabel2.tList)
        except AttributeError:
            raise NotImplementedError()

        return self


    def add(self, t, level):
        """Add a text object to the class object.

        Inputs:
        -----------
        t
            (A matplotlib.Text object) as returned by
            mp.text()

        level
            (int or float) A priority number. If two text labels
            overlap, the one with the lower level is plotted first.

        Returns:
        ------------
        **None**

        """

        assert(isinstance(t, mpl.text.Text))
        t.level = level
        self.tList.append(t)

    def text(self, x, y, text, **kwargs):
        level = kwargs.pop('level', 1)
        self.add(plt.text(x, y, text, **kwargs), level)

    def overlap(self, t1, t2):
        """Do two text objects overlap?

        Internal Function.

        Inputs:
        -------------
        t1, t2
            (matplotlib.Text objects)


        Returns:
        ------------
        boolean
        """
        c1 = getTextBbox(t1)
        c2 = getTextBbox(t2)
        # print(c1, c2, c1.intersection(c1,c2))
        return c1.intersection(c1,c2) is not None

    def render(self):
        """Compute which labels overlap and which ones to hide

        Returns:
        ------------
        **None**
        """
        raise NotImplementedError("Don't call abstract class directly")
    

class SimplePlatLabel(AbstractPlatLabel):
    """Put all the labels to the upper right of their points. 

    Hide lower priority labels that overlap
    """
    def render(self):
        levels = np.array( list(map(lambda x: x.level, self.tList)))

        srt = np.argsort(levels)
        print(srt)
        is_visible = np.ones_like(srt, dtype=bool)

        for i in range(len(srt)):
            t_i = self.tList[ srt[i] ]
            t_i.set_visible(True)

            j = 0
            while j < i:
                print(i, j)
                t_j = self.tList[ srt[j] ]
                if self.overlap(t_i, t_j) and is_visible[j]:
                    print(f"{srt[i]} overlaps with {srt[j]}, {t_j}")
                    print(f"Marking elt {srt[i]} ({t_i}) as invisible")
                    is_visible[i] = False
                    t_i.set_visible(False)
                    break
                j+= 1




class PlatLabel(AbstractPlatLabel):
    """Try different options for ha and va before hiding label"""

    def __init__(self):
        PlatLabel.__init__(self)

    def render(self):

        # Get priority order of labels
        tList = sorted(self.tList, key=lambda x: x.level)
        nLabel = len(tList)

        ha = "left center right".split()
        va = "bottom center top".split()
        placement_options = list(itertools.product(va, ha))

        #Hide all labels to start
        for i in range(nLabel):
            t_i = tList[i]
            t_i.set_visible(False)

        for i in range(nLabel):
            t_i = tList[i]
            #Show this label
            t_i.set_visible(True)

            if i == 0:
                continue 

            #See if we need to hide it again
            # if i == 1:
            #     idebug()
            for pl in placement_options:
                t_i.set_horizontalalignment(pl[1])
                t_i.set_verticalalignment(pl[0])
                # print(  t_i.get_text(), pl, t_i.get_ha(), t_i.get_va())
                t_i.set_visible(True)
                # plt.pause(1)
                # print( getTextBbox(t_i))
                # _x0 = getTextBbox(t_i).x0
                # assert _x0 > 0

                is_displayable = True

                for j in range(i):
                    t_j = tList[j]

                    if  t_j.get_visible() and self.overlap(t_i, t_j):
                        # print(f"{i} overlaps with {j}")
                        is_displayable = False
                        #We found an overlap, this placement is invalid, stop looking
                        #Try the next placement option
                        break

                if is_displayable:
                    #We found a legimiate placement, stop looking for more
                    #Try to place the label i+1
                    t_i.set_visible(True)
                    # plt.pause(.1)
                    # print(t_i)
                    # import ipdb; ipdb.set_trace()
                    break  #Don't try anymore placements
                else:
                    t_i.set_visible(False)




def getTextBbox(text):
    """

    Based on Text.get_window_extent()

    This is called out as a separate function because I expect it will be much more work
    in other plotting toolkits
    """

    visible = text.get_visible()
    text.set_visible(True)
    bbox = text.get_window_extent()
    text.set_visible(visible)
    return bbox


def getPointsPerDataUnit():
    """Compute number of data units per font point size

    Useful if you want to draw text at a particular size relative your data points.
    Naturally the scaled value is invalidated once the axis limits are changed
    or the figure size is changed.

    Returns
    -------
    x_points_per_data, y_points_per_data
        Number of font points per data unit in x and y.


    Example
    --------
    To create text that is as high as 10 data points in the y-axis Do::

        xs, ys = get_points_per_data_unit()
        plt.text(0, 10, "Text", fontsize=10*ys)

    The top of the T in text will align with the "20" position on the y axis

    """
    bbox = plt.gca().get_window_extent().get_points()
    fig = plt.gcf()
    ax = plt.gca()

    bbox = ax.get_window_extent().transformed(fig.dpi_scale_trans.inverted())
    width_inches, height_inches = bbox.width, bbox.height
    width_points = width_inches * fig.dpi
    height_points = height_inches * fig.dpi

    axl = plt.axis()
    yscale_data = axl[3] - axl[2]
    xscale_data = axl[1] - axl[0]

    #This isn't a hard requirement, I just haven't thought through
    #the consequences of -ve values yet.
    # assert np.all(np.array([yscale_points, xscale_points, yscale_data, xscale_data]) > 0)

    x_points_per_data = width_points / xscale_data
    y_points_per_data = height_points / yscale_data * 1
    return x_points_per_data, y_points_per_data


def test1():
    plt.clf()
    plt.axis([-1, 1, -1, 1])
    pl = PlatLabel()
    pl.add(plt.text(0, 0, "Text 0"), 1)
    pl.add(plt.text(.1, 0, "Text 1"), 2)

    assert pl.tList[1].get_visible()
    pl.render()
    assert not pl.tList[1].get_visible()


def test2():
    state = np.random.RandomState(1234)
    plt.clf()
    # plt.axis([0, 1, 0, 1])

    pl = PlatLabel()

    num = 100
    state.rand(10)
    x = state.rand(num)
    y = state.rand(num)
    for i in range(num):
        pl.text(x[i], y[i], f"This is the number {i}", clip_on=True, level=int(i/10))
    pl.render()